/*
    Copyright 2004-2012 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#define _LARGE_FILES        // if it's not supported the tool will work
#define __USE_LARGEFILE64   // without support for large files
#define __USE_FILE_OFFSET64
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS   64

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/stat.h>

#ifdef WIN32
    #include <direct.h>

    #define SYSPATH '\\'
#else
    #include <unistd.h>

    #define stricmp strcasecmp
    #define SYSPATH '/'
    #define __stat64    stat64
    #define _stat64     stat64
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;
typedef uint64_t    u64;

#define PRINTX64(x) (u32)(((x) >> 32) & 0xffffffff), (u32)((x) & 0xffffffff)    // I64x, llx? blah
#if defined(_LARGE_FILES)
    #if defined(__APPLE__)
        #define fseek   fseeko
        #define ftell   ftello
    #elif defined(__FreeBSD__)
    #elif !defined(NOLFS)       // use -DNOLFS if this tool can't be compiled on your OS!
        #define off_t   off64_t
        #define fopen   fopen64
        #define fseek   fseeko64
        #define ftell   ftello64
    #endif
#endif



#define VER         "0.4.1a"
#define PKSIGN      0x04034b50
#define NAMESZ      8192
#define BUFFSZ      32768
#define SHOWNAME    verbose ? (u8 *)"" : name



#pragma pack(2)
typedef struct {
    u32     sign;
    u16     ver;
    u16     flag;
    u16     method;
    u32     timedate;
    u32     crc;
    u32     comp_size;
    u32     uncomp_size;
    u16     name_len;
    u16     extra_len;
    /* filename */
    /* extra */
} zlocal;
#pragma pack()



u8 *show_i64(u64 num);
off_t findzipoff(FILE *fd);
void zipname(FILE *fd, u8 *name, int namelen);
int check_crc(FILE *fdzip, zlocal *local, FILE *fdzip2, off_t fdzip2off);
void b2l_local(zlocal *local);
void b2l_16(u16 *num);
void b2l_32(u32 *num);
u32 crc32(u32 crc, u8 *data, int size);
void std_err(void);



int     verbose = 0,
        endian;



int main(int argc, char *argv[]) {
    FILE    *fd,
            *fd2   = NULL;
    zlocal  local;
    off_t   zoff,
            zoff2  = 0;
    u32     crcok  = 0,
            crcerr = 0,
            nofile = 0;
    int     i;
    static const char   bar[] = "|/-\\";

    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    fputs("\n"
        "ZipComp " VER "\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 2) {
        fprintf(stderr, "\n"
            "Usage: %s <file.ZIP> [directory_or_file.ZIP] [-v]\n"
            "\n", argv[0]);
        exit(1);
    }

    endian = 1;
    if(*(char *)&endian) endian = 0;

    printf("- open zip file:    %s\n", argv[1]);
    fd = fopen(argv[1], "rb");
    if(!fd) std_err();
    zoff = findzipoff(fd);
    fseek(fd, zoff, SEEK_SET);
    if(zoff) printf("- zip offset:       %08x%08x\n", PRINTX64(zoff));

    for(i = 2; i < argc; i++) {
        if(!stricmp(argv[i], "-v")) {
            verbose = 1;
            continue;
        }

        if(chdir(argv[i]) < 0) {
            printf("- open zip file:    %s\n", argv[i]);
            fd2 = fopen(argv[i], "rb");
            if(!fd2) std_err();
            zoff2 = findzipoff(fd2);
            if(zoff2) {
                printf("- zip offset:       %08x%08x\n", PRINTX64(zoff2));
            }
        } else {
            printf("- change directory: %s\n", argv[i]);
        }
    }

    fputs("  Differences:\n\n", stdout);
    while(fread(&local, sizeof(local), 1, fd)) {
        b2l_local(&local);
        if(local.sign != PKSIGN) break;

        fputc(bar[crcok & 3], stderr);
        fputc('\b', stderr);

        switch(check_crc(fd, &local, fd2, zoff2)) {
            case -1: nofile++; break;
            case 0:  crcerr++; break;
            default: crcok++;  break;
        }

        if(fseek(fd, local.comp_size + local.extra_len, SEEK_CUR)) std_err();
    }

    fclose(fd);
    if(fd2) fclose(fd2);

    printf("\n"
        "-------------------------------------\n"
        "%12u unexistent files\n"
        "%12u different files\n"
        "%12u files OK\n"
        "-------------------------------------\n"
        "\n", nofile, crcerr, crcok);

    return(0);
}



u8 *show_i64(u64 num) {
    int     i;
    static u8   nums[22];

    nums[14] = 0;
    for(i = 13; num && (i >= 0); i--) {
        nums[i] = '0' + (num % 10);
        num /= 10;
    }
    return(nums + i + 1);
}



off_t findzipoff(FILE *fd) {
    int     len;
    u32     sign;
    u8      buff[256],
            *p,
            *l;

    fread(buff, 4, 1, fd);
    sign = *(u32 *)buff;
    b2l_32(&sign);
    if(sign == PKSIGN) return(0);

    len = 1;
    for(;;) {
        memmove(buff, buff + len, 3);
        len = fread(buff + 3, 1, sizeof(buff) - 3, fd);
        if(!len) break;

        l = buff + len;
        for(p = buff; p < l; p++) {
            sign = *(u32 *)p;
            b2l_32(&sign);
            if(sign == PKSIGN) return(ftell(fd) - (l - p) - 3);
        }
    }

    return(ftell(fd));
}



void zipname(FILE *fd, u8 *name, int namelen) {
    u8      *l;

    if(namelen > NAMESZ) {
        fprintf(stderr, "Error: the zip contains a too big filename (hacked?)\n\n");
        exit(1);
    }

    if(!fread(name, namelen, 1, fd)) {
        fprintf(stderr, "Error: file is corrupted\n\n");
        exit(1);
    }

    for(l = name + namelen - 1; l >= name; l--) {
        if((*l != '/') && (*l != '\\')) break;
    }
    *(l + 1) = 0;

    for(; l >= name; l--) {
        if((*l == '/') || (*l == '\\')) *l = SYSPATH;
    }
}



int check_crc(FILE *fdzip, zlocal *local, FILE *fdzip2, off_t fdzip2off) {
#ifdef WIN32
    struct  _stati64
#else
    struct  __stat64
#endif
                        xstat;
    FILE    *fd;
    zlocal  local2;
    u32     crc;
    int     len;
    u8      name[NAMESZ + 1],
            name2[NAMESZ + 1],
            buff[BUFFSZ];

    zipname(fdzip, name, local->name_len);
    if(verbose) printf("- %s\n", name);

    if(fdzip2) {
        fseek(fdzip2, fdzip2off, SEEK_SET);
        while(fread(&local2, sizeof(local2), 1, fdzip2)) {
            b2l_local(&local2);
            if(local2.sign != PKSIGN) break;

            zipname(fdzip2, name2, local2.name_len);

            if(!stricmp(name, name2)) {
                crc = local2.crc;
                if(crc != local->crc) {
                    printf("  CRC  %s   %08X instead of %08X\n", SHOWNAME, crc, local->crc);
                    return(0);
                }
                return(1);
            }

            if(fseek(fdzip2, local2.comp_size + local2.extra_len, SEEK_CUR)) break;
        }

        printf("  NONE %s\n", SHOWNAME);
        return(-1);
    }

    if(_stat64(name, &xstat) < 0) {
        printf("  NONE %s\n", SHOWNAME);
        return(-1);
    }
    if(S_ISDIR(xstat.st_mode)) return(1);

    if(xstat.st_size != local->uncomp_size) {
        printf("  SIZE %s   %s, zip %u\n", SHOWNAME, show_i64(xstat.st_size), local->uncomp_size);
        return(0);
    }

    fd = fopen(name, "rb");
    if(!fd) std_err();
    crc = 0xffffffff;
    while((len = fread(buff, 1, BUFFSZ, fd))) {
        crc = crc32(crc, buff, len);
    }
    crc = ~crc;
    fclose(fd);

    if(crc != local->crc) {
        printf("  CRC  %s   %08X instead of %08X\n", SHOWNAME, crc, local->crc);
        return(0);
    }

    return(1);
}



void b2l_local(zlocal *local) {
    if(!endian) return;
    b2l_32(&local->sign);
    b2l_16(&local->ver);
    b2l_16(&local->flag);
    b2l_16(&local->method);
    b2l_32(&local->timedate);
    b2l_32(&local->crc);
    b2l_32(&local->comp_size);
    b2l_32(&local->uncomp_size);
    b2l_16(&local->name_len);
    b2l_16(&local->extra_len);
}



void b2l_16(u16 *num) {
    u16     tmp;

    if(!endian) return;

    tmp = *num;
    *num = ((tmp & 0xff00) >> 8) |
           ((tmp & 0x00ff) << 8);
}



void b2l_32(u32 *num) {
    u32     tmp;

    if(!endian) return;

    tmp = *num;
    *num = ((tmp & 0xff000000) >> 24) |
           ((tmp & 0x00ff0000) >>  8) |
           ((tmp & 0x0000ff00) <<  8) |
           ((tmp & 0x000000ff) << 24);
}



u32 crc32(u32 crc, u8 *data, int size) {
    const static u32	crctable[] = {
        0x00000000, 0x77073096, 0xee0e612c, 0x990951ba,
        0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
        0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
        0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
        0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
        0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
        0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
        0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
        0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
        0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
        0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
        0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
        0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
        0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
        0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
        0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
        0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
        0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
        0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
        0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
        0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
        0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
        0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
        0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
        0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
        0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
        0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
        0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
        0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
        0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
        0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
        0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
        0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
        0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
        0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
        0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
        0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
        0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
        0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
        0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
        0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
        0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
        0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
        0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
        0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
        0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
        0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
        0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
        0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
        0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
        0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
        0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
        0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
        0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
        0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
        0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
        0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
        0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
        0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
        0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
        0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
        0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
        0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
        0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d };

    while(size--) {
        crc = crctable[(*data ^ crc) & 0xff] ^ (crc >> 8);
        data++;
    }
    return(crc);
}



void std_err(void) {
    perror("\nError");
    exit(1);
}


